/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2009, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Sub extension for handling RADIUS Accounting-Request messages */

#define IN_EXTENSION
#define DEFINE_DEBUG_MACRO	sub_auth
#define DECLARE_API_POINTERS
#include <waaad/waaad.h>

#include "rg_common.h"

#ifndef SUB_AUTH_VERBO
#define SUB_AUTH_VERBO 0
#endif /* SUB_AUTH_VERBO */


int sub_auth_verbosity = SUB_AUTH_VERBO;

/* Attributes missing from radius.h */
#define RADIUS_ATTR_CHAP_PASSWORD	3
#define RADIUS_ATTR_ARAP_PASSWORD	70

struct rga_conf_state {
	char * conffile;
	sess_reg_t * sess_hdl;
	void * dl_handle;
	int (*rgw_clients_getkey)(void * cli, unsigned char **key, size_t *key_len);
};

static struct rga_conf_state * auth_conf_parse(char * conffile)
{
	struct rga_conf_state * cs;
	
	TRACE_ENTRY("%p", conffile);
	
	CHECK_MALLOC_DO( cs = malloc(sizeof(struct rga_conf_state)), return NULL );
	memset(cs, 0, sizeof(struct rga_conf_state));
	
	CHECK_FCT_DO( sess_regext( &cs->sess_hdl ), { free(cs); return NULL; } );
	
	cs->conffile = conffile;
	
	if (conffile) {
		TRACE_DEBUG(INFO, "Sub extension Authentication (RFC2865, RFC3579) initialized with configuration: '%s'", cs->conffile);
	} else {
		TRACE_DEBUG(INFO, "Sub extension Authentication (RFC2865, RFC3579) initialized with default configuration");
	}
	
	CHECK_FCT_DO( rg_pointers_init(&cs->dl_handle), return NULL );
	rg_pointers_resolve(cs->rgw_clients_getkey, cs->dl_handle, "rgw_clients_getkey", NULL);
	
	return cs;
}

static void auth_conf_free(struct rga_conf_state * cs)
{
	TRACE_ENTRY("%p", cs);
	CHECK_PARAMS_DO( cs, return );
	CHECK_FCT_DO( sess_deregext( cs->sess_hdl ),  );
	rg_pointers_fini(&cs->dl_handle);
	free(cs);
	return;
}

static int auth_rad_req(struct rga_conf_state * cs, sess_id_t * session, struct radius_msg * rad_req, struct radius_msg ** rad_ans, msg_t ** diam_fw, void * cli )
{
	int idx;
	int got_id = 0;
	int got_mac = 0;
	int got_passwd = 0;
	int got_eap = 0;
	int got_empty_eap = 0;
	uint32_t status_type;
	size_t nattr_used = 0;
	dict_object_t * avp_tun_dict;
	msg_avp_t ** avp_tun = NULL;
	
	TRACE_ENTRY("%p %p %p %p %p %p", cs, session, rad_req, rad_ans, diam_fw, cli);
	CHECK_PARAMS(rad_req && (rad_req->hdr->code == RADIUS_CODE_ACCESS_REQUEST) && rad_ans && diam_fw && *diam_fw);
	
	/* We store the identifier in the session, if any */
	if (session) {
		unsigned char * req_auth;
		CHECK_MALLOC(req_auth = malloc(16));
		memcpy(req_auth, &rad_req->hdr->authenticator[0], 16);
		CHECK_FCT( sess_data_reg(session, cs->sess_hdl, req_auth, free) );
	}
	
	/*
	   Guidelines:
	     http://tools.ietf.org/html/rfc4005#section-9.1
	     http://tools.ietf.org/html/rfc4072#section-6.1

	   When a Translation Agent receives a RADIUS message, the following
	   steps should be taken:

	      -  If a Message-Authenticator attribute is present, the value MUST
        	 be checked but not included in the Diameter message.  If it is
        	 incorrect, the RADIUS message should be silently discarded.
        	 The gateway system SHOULD generate and include a Message-
        	 Authenticator in returned RADIUS responses.
	             -> done in rgw_msg_auth_check

	      -  The transport address of the sender MUST be checked against the
        	 NAS identifying attributes.  See the description of NAS-
        	 Identifier and NAS-IP-Address below.
		     -> done in rgw_clients_check_origin

	      -  The Translation Agent must maintain transaction state
        	 information relevant to the RADIUS request, such as the
        	 Identifier field in the RADIUS header, any existing RADIUS
        	 Proxy-State attribute, and the source IP address and port
        	 number of the UDP packet.  These may be maintained locally in a
        	 state table or saved in a Proxy-Info AVP group.  A Diameter
        	 Session-Id AVP value must be created using a session state
        	 mapping mechanism.
		     -> Identifier, source and port are saved along with the request,
		        and associated with the session state.
		     -> sub_echo_drop should handle the Proxy-State attribute (conf issue)

	      -  If the RADIUS request contained a State attribute and the
        	 prefix of the data is "Diameter/", the data following the
        	 prefix contains the Diameter Origin-Host/Origin-Realm/Session-
        	 Id.  If no such attributes are present and the RADIUS command
        	 is an Access-Request, a new Session-Id is created.  The
        	 Session-Id is included in the Session-Id AVP.
		     -> done in rgw_msg_create_base.

	      -  The Diameter Origin-Host and Origin-Realm AVPs MUST be created
        	 and added by using the information from an FQDN corresponding
        	 to the NAS-IP-Address attribute (preferred if available),
        	 and/or to the NAS-Identifier attribute.  (Note that the RADIUS
        	 NAS-Identifier is not required to be an FQDN.)
		     -> done in rgw_msg_create_base.

	      -  The response MUST have an Origin-AAA-Protocol AVP added,
        	 indicating the protocol of origin of the message.
		     -> what "response" ??? Added to the AAR / DER in this function.

	      -  The Proxy-Info group SHOULD be added, with the local server's
        	 identity specified in the Proxy-Host AVP.  This should ensure
        	 that the response is returned to this system.
		     -> We don't need this, answer is always routed here anyway.
		     
	      For EAP:
	      
	      o  RADIUS EAP-Message attribute(s) are translated to a Diameter
		 EAP-Payload AVP.  If multiple RADIUS EAP-Message attributes are
		 present, they are concatenated and translated to a single Diameter
		 EAP-Payload AVP.
		     -> concatenation done by radius_msg_get_eap

	      -> the remaining is specific conversion rules
	*/
	
	/* Check basic information is there */
	for (idx = 0; idx < rad_req->attr_used; idx++) {
		struct radius_attr_hdr * attr = (struct radius_attr_hdr *)(rad_req->buf + rad_req->attr_pos[idx]);
		switch (attr->type) {
			case RADIUS_ATTR_NAS_IP_ADDRESS:
			case RADIUS_ATTR_NAS_IDENTIFIER:
			case RADIUS_ATTR_NAS_IPV6_ADDRESS:
				got_id = 1;
				break;
			case RADIUS_ATTR_MESSAGE_AUTHENTICATOR:
				got_mac = 1;
				break;
			case RADIUS_ATTR_EAP_MESSAGE:
				got_eap = 1;
				if (attr->length == 2)
					got_empty_eap = 1;
				break;
			case RADIUS_ATTR_USER_PASSWORD:
			case RADIUS_ATTR_CHAP_PASSWORD:
			case RADIUS_ATTR_ARAP_PASSWORD:
				got_passwd += 1;
				break;
		}
	}
	if (!got_id) {
		TRACE_DEBUG(INFO, "RADIUS Access-Request did not contain a NAS IP or Identifier attribute, reject.");
		return EINVAL;
	}
	/* [Note 1] An Access-Request that contains either a User-Password or
	   CHAP-Password or ARAP-Password or one or more EAP-Message attributes
	   MUST NOT contain more than one type of those four attributes.  If it
	   does not contain any of those four attributes, it SHOULD contain a
	   Message-Authenticator.  If any packet type contains an EAP-Message
	   attribute it MUST also contain a Message-Authenticator.  A RADIUS
	   server receiving an Access-Request not containing any of those four
	   attributes and also not containing a Message-Authenticator attribute
	   SHOULD silently discard it.  */
	if (((got_eap + got_passwd) > 1) || (got_eap && !got_mac) || (!got_eap && !got_passwd && !got_mac)) {
		TRACE_DEBUG(INFO, "RADIUS Access-Request not conform to RFC3579 sec 3.3 note 1, discard.");
		return EINVAL;
	}
	
	/* Add the appropriate command code & Auth-Application-Id */
	{
		msg_data_t * header = NULL;
		CHECK_FCT( msg_data ( *diam_fw, &header ) );
		header->msg_flags = CMD_FLAG_REQUEST | CMD_FLAG_PROXIABLE;
		if (got_eap) {
			header->msg_code = 268; /* DER */
			header->msg_appl = 5;   /* Diameter EAP application */
		} else {
			header->msg_code = 265; /* AAR */
			header->msg_appl = 1;   /* Diameter NASREQ */
		}
		
		/* Add the Auth-Application-Id */
		{
			dict_object_t * avp_dict;
			avp_value_t value;
			msg_avp_t * avp;
			CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "Auth-Application-Id", &avp_dict, ENOENT));
			CHECK_FCT( msg_avp_new ( avp_dict, 0, &avp ) );
			value.i32 = header->msg_appl;
			CHECK_FCT( msg_avp_setvalue ( avp, &value ) );
			CHECK_FCT( msg_avp_add ( *diam_fw, MSG_BRW_LAST_CHILD, avp) );
		}
	}
	
	/*  	The type of request is identified through the Auth-Request-Type AVP
		[BASE].  The recommended value for most RADIUS interoperabily
		situations is AUTHORIZE_AUTHENTICATE. */
	
	/* Add Auth-Request-Type AVP */
	{
		dict_object_t * avp_dict;
		avp_value_t value;
		msg_avp_t * avp;
		CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "Auth-Request-Type", &avp_dict, ENOENT));
		CHECK_FCT( msg_avp_new ( avp_dict, 0, &avp ) );
		value.i32 = 3; /* AUTHORIZE_AUTHENTICATE */
		CHECK_FCT( msg_avp_setvalue ( avp, &value ) );
		CHECK_FCT( msg_avp_add ( *diam_fw, MSG_BRW_LAST_CHILD, avp) );
	}
	
	/* Add Origin-AAA-Protocol AVP */
	{
		dict_object_t * avp_dict;
		avp_value_t value;
		msg_avp_t * avp;
		CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "Origin-AAA-Protocol", &avp_dict, ENOENT));
		CHECK_FCT( msg_avp_new ( avp_dict, 0, &avp ) );
		value.i32 = 1; /* RADIUS */
		CHECK_FCT( msg_avp_setvalue ( avp, &value ) );
		CHECK_FCT( msg_avp_add ( *diam_fw, MSG_BRW_LAST_CHILD, avp) );
	}
	
	/* Convert the EAP payload (concat) */
	if (got_eap) {
		dict_object_t * avp_dict;
		avp_value_t value;
		msg_avp_t * avp;
		CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "EAP-Payload", &avp_dict, ENOENT));
		CHECK_FCT( msg_avp_new ( avp_dict, 0, &avp ) );
		
		/*    o  An empty RADIUS EAP-Message attribute (with length 2) signifies
			 EAP-Start, and it is translated to an empty EAP-Payload AVP. */
		if (got_empty_eap) {
			value.os.len = 0;
			value.os.data = "";
		} else {
			CHECK_MALLOC( value.os.data = radius_msg_get_eap(rad_req, &value.os.len) );
		}
		
		CHECK_FCT( msg_avp_setvalue ( avp, &value ) );
		CHECK_FCT( msg_avp_add ( *diam_fw, MSG_BRW_LAST_CHILD, avp) );
	}
	
	/* Tunnel AVPs need some preparation */
	CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "Tunneling", &avp_tun_dict, ENOENT));
	
	/* Convert the attributes one by one */
	for (idx = 0; idx < rad_req->attr_used; idx++) {
		struct radius_attr_hdr * attr = (struct radius_attr_hdr *)(rad_req->buf + rad_req->attr_pos[idx]);
		dict_object_t * avp_dict;
		avp_value_t value;
		msg_avp_t * avp;
		
		switch (attr->type) {
			
			#define CONV_STR( _avp_name_ )	\
				CHECK_PARAMS( attr->length >= 2 );						\
				CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, _avp_name_, &avp_dict, ENOENT));	\
				CHECK_FCT( msg_avp_new ( avp_dict, 0, &avp ) );					\
				value.os.len = attr->length - 2;						\
				value.os.data = (unsigned char *)(attr + 1);					\
				CHECK_FCT( msg_avp_setvalue ( avp, &value ) );					\
				CHECK_FCT( msg_avp_add ( *diam_fw, MSG_BRW_LAST_CHILD, avp) );			\
				
			#define CONV_32B( _avp_name_ )	\
				CHECK_PARAMS( attr->length == 6 );						\
				CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, _avp_name_, &avp_dict, ENOENT));	\
				CHECK_FCT( msg_avp_new ( avp_dict, 0, &avp ) );					\
				{										\
					uint8_t * v = (uint8_t *)(attr + 1);					\
					value.u32  = (v[0] << 24)						\
					           | (v[1] << 16)						\
					           | (v[2] <<  8)						\
					           |  v[3] ;							\
				}										\
				CHECK_FCT( msg_avp_setvalue ( avp, &value ) );					\
				CHECK_FCT( msg_avp_add ( *diam_fw, MSG_BRW_LAST_CHILD, avp) );			\
				
			#define CONV_64B( _avp_name_ )	\
				CHECK_PARAMS( attr->length == 10);						\
				CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, _avp_name_, &avp_dict, ENOENT));	\
				CHECK_FCT( msg_avp_new ( avp_dict, 0, &avp ) );					\
				{										\
					uint8_t * v = (uint8_t *)(attr + 1);					\
					value.u64  = ((uint64_t)(v[0]) << 56)					\
					           | ((uint64_t)(v[1]) << 48)					\
					           | ((uint64_t)(v[2]) << 40)					\
					           | ((uint64_t)(v[3]) << 32)					\
					           | ((uint64_t)(v[4]) << 24)					\
					           | ((uint64_t)(v[5]) << 16)					\
					           | ((uint64_t)(v[6]) <<  8)					\
					           |  (uint64_t)(v[7]) ;					\
				}										\
				CHECK_FCT( msg_avp_setvalue ( avp, &value ) );					\
				CHECK_FCT( msg_avp_add ( *diam_fw, MSG_BRW_LAST_CHILD, avp) );			\
				
		/* RFC 2865 */	
			/*
			      -  The Destination-Realm AVP is created from the information found
        			 in the RADIUS User-Name attribute.
				 	-> done in rgw_msg_create_base
			*/
			case RADIUS_ATTR_USER_NAME:
				CONV_STR( "User-Name" );
				break;
				
			/*
			      -  If the RADIUS User-Password attribute is present, the password
        			 must be unencrypted by using the link's RADIUS shared secret.
        			 The unencrypted value must be forwarded in a User-Password AVP
        			 using Diameter security.
			*/
			case RADIUS_ATTR_USER_PASSWORD:
				if ((attr->length - 2) % 16) {
					TRACE_DEBUG(INFO, "Invalid length of User-Password attribute: %hhd", attr->length);
					return EINVAL;
				}
				{
					/* Decipher following this logic (refers to rfc2865#section-5.2 )
					   b1 = MD5(S + RA)	p1 = c(1) xor b1
					   b2 = MD5(S + c(1))   p2 = c(2) xor b2
					   ...
					*/
					
					uint8_t *ciph = (uint8_t *)(attr+1); 	/* c(i) */
					size_t ciph_len = attr->length - 2;
					uint8_t deciph[128];			/* pi */
					size_t deciph_len = 0;
					uint8_t * secret;			/* S */
					size_t secret_len;
					uint8_t hash[16];			/* b(i) */
					const uint8_t *addr[2];
					size_t len[2];
					
					/* Retrieve the shared secret */
					CHECK_FCT((*cs->rgw_clients_getkey)(cli, &secret, &secret_len));
					
					/* Initial b1 = MD5(S + RA) */
					addr[0] = secret;
					len[0] = secret_len;
					addr[1] = rad_req->hdr->authenticator;
					len[1] = 16;
					md5_vector(2, addr, len, hash);
					
					/* loop */
					while (deciph_len < ciph_len) {
						int i;
						/* pi = c(i) xor bi */
						for (i = 0; i < 16; i++)
							deciph[deciph_len + i] = ciph[deciph_len + i] ^ hash[i];
							/* do we have to remove the padding '\0's ? */
						
						/* b(i+1) = MD5(S + c(i) */
						addr[1] = ciph + deciph_len;
						md5_vector(2, addr, len, hash);
						
						deciph_len += 16;
					}
					
					/* Now save this value in the AVP */
					CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "User-Password", &avp_dict, ENOENT));
					CHECK_FCT( msg_avp_new ( avp_dict, 0, &avp ) );
					value.os.data = &deciph[0];
					value.os.len  = deciph_len;
					CHECK_FCT( msg_avp_setvalue ( avp, &value ) );
					CHECK_FCT( msg_avp_add ( *diam_fw, MSG_BRW_LAST_CHILD, avp) );
				}
				break;
				

			/*
			      -  If the RADIUS CHAP-Password attribute is present, the Ident and
        			 Data portion of the attribute are used to create the CHAP-Auth
        			 grouped AVP.
			*/
			case RADIUS_ATTR_CHAP_PASSWORD:
				CHECK_PARAMS( attr->length == 19 /* RFC 2865 */);
				{
					uint8_t * c = (uint8_t *)(attr + 1);
					msg_avp_t * chap_auth;
					CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "CHAP-Auth", &avp_dict, ENOENT));
					CHECK_FCT( msg_avp_new ( avp_dict, 0, &chap_auth ) );
					CHECK_FCT( msg_avp_add ( *diam_fw, MSG_BRW_LAST_CHILD, chap_auth) );
					
					CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "CHAP-Algorithm", &avp_dict, ENOENT));
					CHECK_FCT( msg_avp_new ( avp_dict, 0, &avp ) );
					value.u32 = 5; /* The only value defined currently... */
					CHECK_FCT( msg_avp_setvalue ( avp, &value ) );
					CHECK_FCT( msg_avp_add ( chap_auth, MSG_BRW_LAST_CHILD, avp) );
					
					CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "CHAP-Ident", &avp_dict, ENOENT));
					CHECK_FCT( msg_avp_new ( avp_dict, 0, &avp ) );
					value.os.data = c;
					value.os.len = 1;
					CHECK_FCT( msg_avp_setvalue ( avp, &value ) );
					CHECK_FCT( msg_avp_add ( chap_auth, MSG_BRW_LAST_CHILD, avp) );
					
					c++;
					
					CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "CHAP-Response", &avp_dict, ENOENT));
					CHECK_FCT( msg_avp_new ( avp_dict, 0, &avp ) );
					value.os.data = c;
					value.os.len = attr->length - 3;
					CHECK_FCT( msg_avp_setvalue ( avp, &value ) );
					CHECK_FCT( msg_avp_add ( chap_auth, MSG_BRW_LAST_CHILD, avp) );
				}
				break;
				
			case RADIUS_ATTR_NAS_IP_ADDRESS:
				CONV_STR( "NAS-IP-Address" );
				break;
				
			case RADIUS_ATTR_NAS_PORT:
				CONV_32B( "NAS-Port" );
				break;
				
			case RADIUS_ATTR_SERVICE_TYPE:
				CONV_32B( "Service-Type" );
				break;
				
			case RADIUS_ATTR_FRAMED_PROTOCOL:
				CONV_32B( "Framed-Protocol" );
				break;
				
			case RADIUS_ATTR_FRAMED_IP_ADDRESS:
				CONV_STR( "Framed-IP-Address" );
				break;
				
			case RADIUS_ATTR_FRAMED_IP_NETMASK:
				CONV_STR( "Framed-IP-Netmask" );
				break;
				
			/* Framed-Routing never present in an Access-Request */
			/* Filter-Id never present in an Access-Request */
				
			case RADIUS_ATTR_FRAMED_MTU:
				CONV_32B( "Framed-MTU" );
				break;
			
			case RADIUS_ATTR_FRAMED_COMPRESSION:
				CONV_32B( "Framed-Compression" );
				break;
			
			case RADIUS_ATTR_LOGIN_IP_HOST:
				CONV_STR( "Login-IP-Host" );
				break;
				
			/* Login-Service never present in an Access-Request */
			/* Login-TCP-Port never present in an Access-Request */
			/* Reply-Message never present in an Access-Request */
			
			case RADIUS_ATTR_CALLBACK_NUMBER:
				CONV_STR( "Callback-Number" );
				break;
				
			/* Callback-Id never present in an Access-Request */
			/* Framed-Route never present in an Access-Request */
			/* Framed-IPX-Network never present in an Access-Request */
				
			case RADIUS_ATTR_STATE:
				CONV_STR( "State" );
				break;
			
			/* Class never present in an Access-Request */
				
			case RADIUS_ATTR_VENDOR_SPECIFIC:
				/* RFC 4005, Section 9.6 : 
					   Systems that don't have vendor format knowledge MAY discard such
					   attributes without knowing a suitable translation.
					   
					   [conversion rule in 9.6.2]
				 */
				if (attr->length >= 6) {
					uint32_t vendor_id;
					uint8_t * c = (uint8_t *)(attr + 1);
					
					vendor_id = c[0] << 24 | c[1] << 16 | c[2] << 8 | c[3];
					c += 4;
					
					switch (vendor_id) {
						
						/* For the vendors we KNOW they follow the VSA recommended format, we convert following the rules of RFC4005 (9.6.2) */
						case RADIUS_VENDOR_ID_MICROSOFT : /* RFC 2548 */
						/* other vendors ? */
						{
							size_t left;
							struct radius_attr_vendor *vtlv;
							
							CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "waaad_generic_os", &avp_dict, ENOENT));
							
							left = attr->length - 6;
							vtlv = (struct radius_attr_vendor *)c;
						
							while ((left >= 2) && (vtlv->vendor_length <= left)) {
								msg_avp_data_t * avp_hdr;
								
								CHECK_FCT( msg_avp_new ( avp_dict, 0, &avp ) );
								value.os.len = vtlv->vendor_length - 2;
								value.os.data = (unsigned char *)(vtlv + 1);
								CHECK_FCT( msg_avp_setvalue ( avp, &value ) );
								
								CHECK_FCT( msg_avp_data( avp, &avp_hdr ) );
								avp_hdr->avp_code = vtlv->vendor_type;
								avp_hdr->avp_flags = AVP_FLAG_VENDOR;
								/* avp_hdr->avp_len = ; */
								avp_hdr->avp_vendor = vendor_id;
								
								CHECK_FCT( msg_avp_add ( *diam_fw, MSG_BRW_LAST_CHILD, avp) );
								
								c += vtlv->vendor_length;
								left -= vtlv->vendor_length;
								vtlv = (struct radius_attr_vendor *)c;
							}
						}
						break;
						
						/* Other vendors we KNOw how to convert the attributes would be added here... */
						/* case RADIUS_VENDOR_ID_CISCO :
							break; */
						/* case RADIUS_VENDOR_ID_IETF : (extended RADIUS attributes)
							break; */
						
						/* When we don't know, just discard the attribute... VSA are optional with regards to RADIUS anyway */
						default:
							/* do nothing */
							TRACE_DEBUG(FULL, "VSA attribute from vendor %d discarded", vendor_id);
							
					}
				}
				break;
				
			/* Session-Timeout never present in an Access-Request */
			/* Idle-Timeout never present in an Access-Request */
			/* Termination-Action never present in an Access-Request */
				
			case RADIUS_ATTR_CALLED_STATION_ID:
				CONV_STR( "Called-Station-Id" );
				break;
			
			case RADIUS_ATTR_CALLING_STATION_ID:
				CONV_STR( "Calling-Station-Id" );
				break;
			
			case RADIUS_ATTR_NAS_IDENTIFIER:
				CONV_STR( "NAS-Identifier" );
				break;
			
			/* Proxy-State is supposed to be handled by sub_echo_drop, we don't touch it here */
			
			case RADIUS_ATTR_LOGIN_LAT_SERVICE:
				CONV_STR( "Login-LAT-Service" );
				break;
			
			case RADIUS_ATTR_LOGIN_LAT_NODE:
				CONV_STR( "Login-LAT-Node" );
				break;
			
			case RADIUS_ATTR_LOGIN_LAT_GROUP:
				CONV_STR( "Login-LAT-Group" );
				break;
			
			/* Framed-AppleTalk-Link never present in an Access-Request */
			/* Framed-AppleTalk-Network never present in an Access-Request */
			/* Framed-AppleTalk-Zone never present in an Access-Request */
			
			case RADIUS_ATTR_CHAP_CHALLENGE:
				CONV_STR( "CHAP-Challenge" );
				break;
			
			case RADIUS_ATTR_NAS_PORT_TYPE:
				CONV_32B( "NAS-Port-Type" );
				break;
			
			case RADIUS_ATTR_PORT_LIMIT:
				CONV_32B( "Port-Limit" );
				break;
			
			case RADIUS_ATTR_LOGIN_LAT_PORT:
				CONV_STR( "Login-LAT-Port" );
				break;
			
			
		/* RFC 3162 */	
			case RADIUS_ATTR_NAS_IPV6_ADDRESS:
				CONV_STR( "NAS-IPv6-Address" );
				break;
				
			case RADIUS_ATTR_FRAMED_INTERFACE_ID:
				CONV_64B( "Framed-Interface-Id" );
				break;
				
			case RADIUS_ATTR_FRAMED_IPV6_PREFIX:
				CONV_STR( "Framed-IPv6-Prefix" );
				break;
				
			case RADIUS_ATTR_LOGIN_IPV6_HOST:
				CONV_STR( "Login-IPv6-Host" );
				break;
				
			/* Framed-IPv6-Route never present in an Access-Request */
			/* Framed-IPv6-Pool never present in an Access-Request */


		/* RFC 2868 */	
			#define AVP_TUN_PREPARE() {										\
						if (avp_tun == NULL) {								\
							CHECK_MALLOC( avp_tun = malloc(sizeof( msg_avp_t * ) * 32 ) );		\
							memset(avp_tun, 0, sizeof( msg_avp_t * )* 32);				\
						}										\
						tag = *(uint8_t *)(attr + 1);							\
						if (tag > 0x1F) tag = 0;							\
						if (avp_tun[tag] == NULL) {							\
							CHECK_FCT( msg_avp_new ( avp_tun_dict, 0, &avp_tun[tag] ) );		\
							CHECK_FCT( msg_avp_add ( *diam_fw, MSG_BRW_LAST_CHILD, avp_tun[tag]) );	\
						}										\
					}
					
			#define CONV_TUN_STR( _avp_name_ ) {							\
				uint8_t tag;									\
				AVP_TUN_PREPARE();								\
				CHECK_PARAMS( attr->length >= 3);						\
				CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, _avp_name_, &avp_dict, ENOENT));	\
				CHECK_FCT( msg_avp_new ( avp_dict, 0, &avp ) );					\
				value.os.len = attr->length - (tag ? 3 : 2);					\
				value.os.data = ((unsigned char *)(attr + 1)) + (tag ? 1 : 0);			\
				CHECK_FCT( msg_avp_setvalue ( avp, &value ) );					\
				CHECK_FCT( msg_avp_add ( avp_tun[tag], MSG_BRW_LAST_CHILD, avp) );		\
				}
				
			#define CONV_TUN_24B( _avp_name_ ) {							\
				uint8_t tag;									\
				AVP_TUN_PREPARE();								\
				CHECK_PARAMS( attr->length == 6);						\
				CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, _avp_name_, &avp_dict, ENOENT));	\
				CHECK_FCT( msg_avp_new ( avp_dict, 0, &avp ) );					\
				{										\
					uint8_t * v = (uint8_t *)(attr + 1);					\
					value.u32 = (v[1] << 16) | (v[2] <<8) | v[3] ;				\
				}										\
				CHECK_FCT( msg_avp_setvalue ( avp, &value ) );					\
				CHECK_FCT( msg_avp_add ( avp_tun[tag], MSG_BRW_LAST_CHILD, avp) );		\
				}

			/*
			      -  If the RADIUS message contains Tunnel information [RADTunnels],
        			 the attributes or tagged groups should each be converted to a
        			 Diameter Tunneling Grouped AVP set.  If the tunnel information
        			 contains a Tunnel-Password attribute, the RADIUS encryption
        			 must be resolved, and the password forwarded, by using Diameter
        			 security methods.
				    -> If the RADIUS message does not use properly the Tag info, result is unpredictible here.. 
			*/
			case RADIUS_ATTR_TUNNEL_TYPE:
				CONV_TUN_24B( "Tunnel-Type" );
				break;
			
			case RADIUS_ATTR_TUNNEL_MEDIUM_TYPE:
				CONV_TUN_24B( "Tunnel-Medium-Type" );
				break;
			
			case RADIUS_ATTR_TUNNEL_CLIENT_ENDPOINT:
				CONV_TUN_STR( "Tunnel-Client-Endpoint" );
				break;
			
			case RADIUS_ATTR_TUNNEL_SERVER_ENDPOINT:
				CONV_TUN_STR( "Tunnel-Server-Endpoint" );
				break;
			
			/* Tunnel-Password never present in an Access-Request */

			case RADIUS_ATTR_TUNNEL_PRIVATE_GROUP_ID:
				CONV_TUN_STR( "Tunnel-Private-Group-Id" );
				break;
			
			/* Tunnel-Assignment-ID never present in an Access-Request */
			
			case RADIUS_ATTR_TUNNEL_PREFERENCE:
				CONV_TUN_24B( "Tunnel-Preference" );
				break;
			
			case RADIUS_ATTR_TUNNEL_CLIENT_AUTH_ID:
				CONV_TUN_STR( "Tunnel-Client-Auth-Id" );
				break;
			
			case RADIUS_ATTR_TUNNEL_SERVER_AUTH_ID:
				CONV_TUN_STR( "Tunnel-Server-Auth-Id" );
				break;
			
			
		/* RFC 2869 */	
			case RADIUS_ATTR_ARAP_PASSWORD:
				CONV_STR( "ARAP-Password" );
				break;
				
			/* ARAP-Features never present in an Access-Request */
			/* ARAP-Zone-Access never present in an Access-Request */
			
			case RADIUS_ATTR_ARAP_SECURITY:
				CONV_32B( "ARAP-Security" );
				break;
			
			case RADIUS_ATTR_ARAP_SECURITY_DATA:
				CONV_STR( "ARAP-Security-Data" );
				break;
			
			/* Password-Retry never present in an Access-Request */
			/* Prompt never present in an Access-Request */
			
			case RADIUS_ATTR_CONNECT_INFO:
				CONV_STR( "Connect-Info" );
				break;
			
			/* Configuration-Token never present in an Access-Request */
			/* ARAP-Challenge-Response never present in an Access-Request */
			/* Acct-Interim-Interval never present in an Access-Request */
			
			case RADIUS_ATTR_NAS_PORT_ID:
				CONV_STR( "NAS-Port-Id" );
				break;
			
			/* Framed-Pool never present in an Access-Request */
			
				
		/* RFC 2869 / 3579 */	
			case RADIUS_ATTR_ORIGINATING_LINE_INFO:
				CONV_STR( "Originating-Line-Info" );
				break;
				
			case RADIUS_ATTR_MESSAGE_AUTHENTICATOR:
			case RADIUS_ATTR_EAP_MESSAGE:
				/* It was already handled, just remove the attribute */
				break;
				
		/* Default */		
			default: /* unknown attribute */
				/* We just keep the attribute in the RADIUS message */
				rad_req->attr_pos[nattr_used++] = rad_req->attr_pos[idx];
		}
	}
	
	/* Destroy tunnel pointers (if we used it) */
	free(avp_tun);
	
	/* Update the radius message to remove all handled attributes */
	rad_req->attr_used = nattr_used;

	return 0;
}

static int auth_diam_ans(struct rga_conf_state * cs, sess_id_t * session, msg_t ** diam_ans, struct radius_msg ** rad_fw, void * cli )
{
	msg_data_t *mdata;
	msg_avp_t *avp, *next, *avp_x, *avp_y, *asid, *aoh;
	msg_avp_data_t *adata, *sid, *oh;
	dict_object_t * avp_dict;
	char buf[254]; /* to store some attributes values (with final '\0') */
	int ta_set = 0;
	uint8_t	tuntag = 0;
	unsigned char * req_auth = NULL;
	
	TRACE_ENTRY("%p %p %p %p %p", cs, session, diam_ans, rad_fw, cli);
	CHECK_PARAMS(cs && session && diam_ans && *diam_ans && rad_fw && *rad_fw);
	
	if (session) {
		int ret = sess_data_dereg( session, cs->sess_hdl, (void *)&req_auth );
		if (ret == ENOENT) {
			TRACE_DEBUG(FULL, "No data saved in the session");
		} else {
			CHECK_FCT(ret); /* Return if another error occurred */
		}
	}
	
	CHECK_FCT( msg_data( *diam_ans, &mdata ) );
	
	/*	
	      -  If the Diameter Command-Code is set to AA-Answer and the
        	 Result-Code AVP is set to DIAMETER_MULTI_ROUND_AUTH, the
        	 gateway must send a RADIUS Access-Challenge.  This must have
        	 the Origin-Host, Origin-Realm, and Diameter Session-Id AVPs
        	 encapsulated in the RADIUS State attribute, with the prefix
        	 "Diameter/", concatenated in the above order separated with "/"
        	 characters, in UTF-8 [UTF-8].  This is necessary to ensure that
        	 the Translation Agent receiving the subsequent RADIUS Access-
        	 Request will have access to the Session Identifier and be able
        	 to set the Destination-Host to the correct value.
		 	-> done here bellow
		 
	      -  If the Command-Code is set to AA-Answer, the Diameter Session-
        	 Id AVP is saved in a new RADIUS Class attribute whose format
        	 consists of the string "Diameter/" followed by the Diameter
        	 Session Identifier.  This will ensure that the subsequent
        	 Accounting messages, which could be received by any Translation
        	 Agent, would have access to the original Diameter Session
        	 Identifier.
		 	-> done here but only for Access-Accept messages (Result-Code = success)
	 */

	/* _trunc_ = 0 => error; _trunc_ = 1 => truncate; _trunc = 2 => create several attributes */
#define CONV2_STR( _attr_, _data_, _len_, _trunc_)	{					\
	size_t __l = (size_t)(_len_);								\
	size_t __off = 0;									\
	TRACE_DEBUG(FULL, "Converting attribute to "#_attr_);					\
	if ((_trunc_) == 0) {									\
		CHECK_PARAMS( __l <= 253 );							\
	}											\
	if ((__l > 253) && (_trunc_ == 1)) {							\
		TRACE_DEBUG(FULL, "Attribute truncated!");					\
		__l = 253;									\
	}											\
	do {											\
		size_t __w = (__l > 253) ? 253 : __l;						\
		CHECK_MALLOC(radius_msg_add_attr(*rad_fw, (_attr_), (_data_) + __off, __w));	\
		__off += __w;									\
		__l   -= __w;									\
	} while (__l);										\
}

#define CONV2_32B( _attr_, _data_)	{							\
	uint32_t __v = htonl((uint32_t)(_data_));						\
	TRACE_DEBUG(FULL, "Converting attribute to "#_attr_);					\
	CHECK_MALLOC(radius_msg_add_attr(*rad_fw, (_attr_), (uint8_t *)&__v, sizeof(__v)));	\
}

#define CONV2_64B( _attr_, _data_)	{	/* Some systems do not provide htonll */	\
	uint64_t __v = ((uint64_t)htonl((uint32_t)(((uint64_t)(_data_)) >> 32))) << 32;		\
		__v  |= htonl((uint32_t)(_data_));						\
	TRACE_DEBUG(FULL, "Converting attribute to "#_attr_);					\
	CHECK_MALLOC(radius_msg_add_attr(*rad_fw, (_attr_), (uint8_t *)&__v, sizeof(__v)));	\
}
	/* Search the different AVPs we handle here */
	CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "Session-Id", &avp_dict, ENOENT));
	CHECK_FCT( msg_search_avp (*diam_ans, avp_dict, &asid) );
	CHECK_FCT( msg_avp_data ( asid, &sid ) );
	CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "Origin-Host", &avp_dict, ENOENT));
	CHECK_FCT( msg_search_avp (*diam_ans, avp_dict, &aoh) );
	CHECK_FCT( msg_avp_data ( aoh, &oh ) );

	/* Check the Diameter error code */
	CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "Result-Code", &avp_dict, ENOENT));
	CHECK_FCT( msg_search_avp (*diam_ans, avp_dict, &avp) );
	CHECK_FCT( msg_avp_data ( avp, &adata ) );
	switch (adata->avp_data->u32) {
		case 1001: /* DIAMETER_MULTI_ROUND_AUTH */
			(*rad_fw)->hdr->code = RADIUS_CODE_ACCESS_CHALLENGE;
			break;
		case 2001: /* DIAMETER_SUCCESS */
		case 2002: /* DIAMETER_LIMITED_SUCCESS */
			(*rad_fw)->hdr->code = RADIUS_CODE_ACCESS_ACCEPT;
			break;
		
		default:
			(*rad_fw)->hdr->code = RADIUS_CODE_ACCESS_REJECT;
			log_error("Received Diameter answer with error code '%d' from server '%.*s', session %.*s, sending Access-Reject\n",
					adata->avp_data->u32, 
					oh->avp_data->os.len, oh->avp_data->os.data,
					sid->avp_data->os.len, sid->avp_data->os.data);
			CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "Error-Message", &avp_dict, ENOENT));
			CHECK_FCT( msg_search_avp (*diam_ans, avp_dict, &avp) );
			if (avp) {
				CHECK_FCT( msg_avp_data ( avp, &adata ) );
				log_error("  Error-Message content: '%.*s'\n",
						adata->avp_data->os.len, adata->avp_data->os.data);
			}
			CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "Error-Reporting-Host", &avp_dict, ENOENT));
			CHECK_FCT( msg_search_avp (*diam_ans, avp_dict, &avp) );
			if (avp) {
				CHECK_FCT( msg_avp_data ( avp, &adata ) );
				log_error("  Error-Reporting-Host: '%.*s'\n",
						adata->avp_data->os.len, adata->avp_data->os.data);
			}
			CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "Failed-AVP", &avp_dict, ENOENT));
			CHECK_FCT( msg_search_avp (*diam_ans, avp_dict, &avp) );
			if (avp) {
				log_error("  Failed-AVP was present in the message\n");
				/* Dump its content ? */
			}
			return 0;
	}
	/* Remove this Result-Code avp */
	CHECK_FCT( msg_free( avp, 1 ) );
	
	/* Process creation of the State or Class attribute with session information */
	CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "Origin-Realm", &avp_dict, ENOENT));
	CHECK_FCT( msg_search_avp (*diam_ans, avp_dict, &avp) );
	CHECK_FCT( msg_avp_data ( avp, &adata ) );
	
	/* Now, save the session-id and eventually server info in a STATE or CLASS attribute */
	if ((*rad_fw)->hdr->code == RADIUS_CODE_ACCESS_CHALLENGE) {
		if (sizeof(buf) < snprintf(buf, sizeof(buf), "Diameter/%.*s/%.*s/%.*s", 
				oh->avp_data->os.len,  oh->avp_data->os.data,
				adata->avp_data->os.len,  adata->avp_data->os.data,
				sid->avp_data->os.len, sid->avp_data->os.data)) {
			TRACE_DEBUG(INFO, "Data truncated in State attribute: %s", buf);
		}
		CONV2_STR(RADIUS_ATTR_STATE, buf, strlen(buf), 0);
	}
	/* The RFC text says that this should always be the case, but it seems odd... */
	if ((*rad_fw)->hdr->code == RADIUS_CODE_ACCESS_ACCEPT) {
		if (sizeof(buf) < snprintf(buf, sizeof(buf), "Diameter/%.*s", 
				sid->avp_data->os.len, sid->avp_data->os.data)) {
			TRACE_DEBUG(INFO, "Data truncated in Class attribute: %s", buf);
		}
		CONV2_STR(RADIUS_ATTR_CLASS, buf, strlen(buf), 0);
	}
	
	/* Unlink the Origin-Realm now; the others are unlinked at the end of this function */
	CHECK_FCT( msg_free( avp, 1 ) );
	
	CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "Session-Timeout", &avp_dict, ENOENT));
	CHECK_FCT( msg_search_avp (*diam_ans, avp_dict, &avp) );
	CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "Authorization-Lifetime", &avp_dict, ENOENT));
	CHECK_FCT( msg_search_avp (*diam_ans, avp_dict, &avp_x) );
	CHECK_FCT( dict_search( DICT_AVP, AVP_BY_NAME, "Re-Auth-Request-Type", &avp_dict, ENOENT));
	CHECK_FCT( msg_search_avp (*diam_ans, avp_dict, &avp_y) );
	/*	
	   When translating a Diameter AA-Answer (with successful result code)
	   to RADIUS Access-Accept that contains a Session-Timeout or
	   Authorization-Lifetime AVP, take the following steps:
	   
	      -  If the Diameter message contains a Session-Timeout AVP but no
        	 Authorization-Lifetime AVP, translate it to a Session-Timeout
        	 attribute (not a Termination-Action).
	*/
	if ((avp != NULL) && (avp_x == NULL)) {
		CHECK_FCT( msg_avp_data ( avp, &adata ) );
		CONV2_32B( RADIUS_ATTR_SESSION_TIMEOUT, adata->avp_data->u32 );
	}
	
	/*	
	      -  If the Diameter message contains an Authorization-Lifetime AVP
        	 but no Session-Timeout AVP, translate it to a Session-Timeout
        	 attribute and a Termination-Action set to AA-REQUEST.  (Remove
        	 Authorization-Lifetime and Re-Auth-Request-Type.)
	*/
	if ((avp == NULL) && (avp_x != NULL)) {
		CHECK_FCT( msg_avp_data ( avp_x, &adata ) );
		CONV2_32B( RADIUS_ATTR_SESSION_TIMEOUT, adata->avp_data->u32 );
		CONV2_32B( RADIUS_ATTR_TERMINATION_ACTION, RADIUS_TERMINATION_ACTION_RADIUS_REQUEST );
		ta_set = 1;
	}
		
	/*	
	      -  If the Diameter message has both, the Session-Timeout must be
        	 greater than or equal to the Authorization-Lifetime (required
        	 by [BASE]).  Translate it to a Session-Timeout value (with
        	 value from Authorization-Lifetime AVP, the smaller one) and
        	 with the Termination-Action set to AA-REQUEST.  (Remove the
        	 Authorization-Lifetime and Re-Auth-Request-Type.)
	*/
	if ((avp != NULL) && (avp_x != NULL)) {
		CHECK_FCT( msg_avp_data ( avp_x, &adata ) );
		CONV2_32B( RADIUS_ATTR_SESSION_TIMEOUT, adata->avp_data->u32 );
		CONV2_32B( RADIUS_ATTR_TERMINATION_ACTION, RADIUS_TERMINATION_ACTION_RADIUS_REQUEST );
		ta_set = 1;
	}
	
	/*  -> Not too sure about Auth-Grace-Period... we'll just discard it for now */
	
	if (avp) {
		CHECK_FCT( msg_free( avp, 1 ) );
	}
	if (avp_x) {
		CHECK_FCT( msg_free( avp_x, 1 ) );
	}
	if (avp_y) {
		CHECK_FCT( msg_free( avp_y, 1 ) );
	}
	
	
	/*
	      -  If a Proxy-State attribute was present in the RADIUS request,
        	 the same attribute is added in the response.  This information
        	 may be found in the Proxy-Info AVP group, or in a local state
        	 table.
		 	-> handled by sub_echo_drop

	      -  If state information regarding the RADIUS request was saved in
        	 a Proxy-Info AVP or local state table, the RADIUS Identifier
        	 and UDP IP Address and port number are extracted and used in
        	 issuing the RADIUS reply.
		 	-> was saved with the full request
	*/
	
	
	/* Now loop in the list of AVPs and convert those that we know how */
	CHECK_FCT( msg_browse(*diam_ans, MSG_BRW_FIRST_CHILD, &next, NULL) );
	
	while (next) {
		int handled = 1;
		avp = next;
		CHECK_FCT( msg_browse(avp, MSG_BRW_NEXT, &next, NULL) );
		
		CHECK_FCT( msg_avp_data ( avp, &adata ) );
		
		if (!(adata->avp_flags & AVP_FLAG_VENDOR)) {
			switch (adata->avp_code) {
				
		/* RFC 4005 (AVP in the order of the AA-Request/Answer AVP Table) */
				case DIAM_ATTR_ACCT_INTERIM_INTERVAL:
					CONV2_32B(RADIUS_ATTR_ACCT_INTERIM_INTERVAL, adata->avp_data->u32);
					break;
					
				case DIAM_ATTR_ARAP_CHALLENGE_RESPONSE:
					CONV2_STR(RADIUS_ATTR_ARAP_CHALLENGE_RESPONSE, adata->avp_data->os.data, adata->avp_data->os.len, 0);
					break;
					
				case DIAM_ATTR_ARAP_FEATURES:
					CONV2_STR(RADIUS_ATTR_ARAP_FEATURES, adata->avp_data->os.data, adata->avp_data->os.len, 0);
					break;
					
				/* ARAP-Password is not present in answers */
					
				case DIAM_ATTR_ARAP_SECURITY:
					CONV2_32B(RADIUS_ATTR_ARAP_SECURITY, adata->avp_data->u32);
					break;
					
				case DIAM_ATTR_ARAP_SECURITY_DATA:
					CONV2_STR(RADIUS_ATTR_ARAP_SECURITY_DATA, adata->avp_data->os.data, adata->avp_data->os.len, 2);
					break;
					
				case DIAM_ATTR_ARAP_ZONE_ACCESS:
					CONV2_32B(RADIUS_ATTR_ARAP_ZONE_ACCESS, adata->avp_data->u32);
					break;
					
				case DIAM_ATTR_AUTH_APPLICATION_ID:
					/* We just remove this AVP */
					break;
					
				case DIAM_ATTR_AUTH_GRACE_PERIOD:
					/* We just remove this AVP (?) */
					break;
				
				case DIAM_ATTR_AUTH_REQUEST_TYPE:
					/* We only check the value */
					if (adata->avp_data->u32 != 3) {
						log_normal("Received Diameter answer with Auth-Request-Type set to %d (%s) from server %.*s, session %.*s.\n"
								"  This may cause interoperability with RADIUS.\n",
								adata->avp_data->u32,
								(adata->avp_data->u32 == 1) ? "AUTHENTICATE_ONLY" :
									((adata->avp_data->u32 == 2) ? "AUTHORIZE_ONLY" : "???"),
								oh->avp_data->os.len, oh->avp_data->os.data, 
								sid->avp_data->os.len, sid->avp_data->os.len);
					}
					break;
				
				case DIAM_ATTR_AUTH_SESSION_STATE:
					if ((!ta_set) && (adata->avp_data->u32 == 0 /* STATE_MAINTAINED */)) {
						CONV2_32B( RADIUS_ATTR_TERMINATION_ACTION, RADIUS_TERMINATION_ACTION_RADIUS_REQUEST );
					}
					break;
					
				/* Authorization-Lifetime already handled */
				
				case DIAM_ATTR_CALLBACK_ID:
					CONV2_STR(RADIUS_ATTR_CALLBACK_ID, adata->avp_data->os.data, adata->avp_data->os.len, 1);
					break;
				
				case DIAM_ATTR_CALLBACK_NUMBER:
					CONV2_STR(RADIUS_ATTR_CALLBACK_NUMBER, adata->avp_data->os.data, adata->avp_data->os.len, 1);
					break;
				
				/* Called-Station-Id is not present in answers */
				/* Calling-Station-Id is not present in answers */
				/* CHAP-Auth is not present in answers */
				/* CHAP-Challenge is not present in answers */
					
				case DIAM_ATTR_CLASS:
					CONV2_STR(RADIUS_ATTR_CLASS, adata->avp_data->os.data, adata->avp_data->os.len, 2);
					break;
				
				case DIAM_ATTR_CONFIGURATION_TOKEN:
					/* We might as well remove it since it's not supposed to be sent to the NAS... */
					CONV2_STR(RADIUS_ATTR_CONFIGURATION_TOKEN, adata->avp_data->os.data, adata->avp_data->os.len, 2);
					break;
				
				/* Connect-Info is not present in answers */
				
				case DIAM_ATTR_FILTER_ID:
					CONV2_STR(RADIUS_ATTR_FILTER_ID, adata->avp_data->os.data, adata->avp_data->os.len, 2);
					break;
					
				case DIAM_ATTR_FRAMED_APPLETALK_LINK:
					CONV2_32B(RADIUS_ATTR_FRAMED_APPLETALK_LINK, adata->avp_data->u32);
					break;
					
				case DIAM_ATTR_FRAMED_APPLETALK_NETWORK:
					CONV2_32B(RADIUS_ATTR_FRAMED_APPLETALK_NETWORK, adata->avp_data->u32);
					break;
					
				case DIAM_ATTR_FRAMED_APPLETALK_ZONE:
					CONV2_STR(RADIUS_ATTR_FRAMED_APPLETALK_ZONE, adata->avp_data->os.data, adata->avp_data->os.len, 1);
					break;
					
				case DIAM_ATTR_FRAMED_COMPRESSION:
					CONV2_32B(RADIUS_ATTR_FRAMED_COMPRESSION, adata->avp_data->u32);
					break;
					
				case DIAM_ATTR_FRAMED_INTERFACE_ID:
					CONV2_64B(RADIUS_ATTR_FRAMED_INTERFACE_ID, adata->avp_data->u64);
					break;
					
				case DIAM_ATTR_FRAMED_IP_ADDRESS:
					CONV2_STR(RADIUS_ATTR_FRAMED_IP_ADDRESS, adata->avp_data->os.data, adata->avp_data->os.len, 0);
					break;
					
				case DIAM_ATTR_FRAMED_IP_NETMASK:
					CONV2_STR(RADIUS_ATTR_FRAMED_IP_NETMASK, adata->avp_data->os.data, adata->avp_data->os.len, 0);
					break;
					
				case DIAM_ATTR_FRAMED_IPV6_PREFIX:
					CONV2_STR(RADIUS_ATTR_FRAMED_IPV6_PREFIX, adata->avp_data->os.data, adata->avp_data->os.len, 0);
					break;
					
				case DIAM_ATTR_FRAMED_IPV6_POOL:
					CONV2_STR(RADIUS_ATTR_FRAMED_IPV6_POOL, adata->avp_data->os.data, adata->avp_data->os.len, 1);
					break;
					
				case DIAM_ATTR_FRAMED_IPV6_ROUTE:
					CONV2_STR(RADIUS_ATTR_FRAMED_IPV6_ROUTE, adata->avp_data->os.data, adata->avp_data->os.len, 1);
					break;
					
				case DIAM_ATTR_FRAMED_IPX_NETWORK:
					CONV2_32B(RADIUS_ATTR_FRAMED_IPX_NETWORK, adata->avp_data->u32);
					break;
					
				case DIAM_ATTR_FRAMED_MTU:
					CONV2_32B(RADIUS_ATTR_FRAMED_MTU, adata->avp_data->u32);
					break;
					
				case DIAM_ATTR_FRAMED_POOL:
					CONV2_STR(RADIUS_ATTR_FRAMED_POOL, adata->avp_data->os.data, adata->avp_data->os.len, 1);
					break;
					
				case DIAM_ATTR_FRAMED_PROTOCOL:
					CONV2_32B(RADIUS_ATTR_FRAMED_PROTOCOL, adata->avp_data->u32);
					break;
					
				case DIAM_ATTR_FRAMED_ROUTE:
					CONV2_STR(RADIUS_ATTR_FRAMED_ROUTE, adata->avp_data->os.data, adata->avp_data->os.len, 1);
					break;
					
				case DIAM_ATTR_FRAMED_ROUTING:
					CONV2_32B(RADIUS_ATTR_FRAMED_ROUTING, adata->avp_data->u32);
					break;
					
				case DIAM_ATTR_IDLE_TIMEOUT:
					CONV2_32B(RADIUS_ATTR_IDLE_TIMEOUT, adata->avp_data->u32);
					break;
					
				case DIAM_ATTR_LOGIN_IP_HOST:
					CONV2_STR(RADIUS_ATTR_LOGIN_IP_HOST, adata->avp_data->os.data, adata->avp_data->os.len, 0);
					break;
					
				case DIAM_ATTR_LOGIN_IPV6_HOST:
					CONV2_STR(RADIUS_ATTR_LOGIN_IPV6_HOST, adata->avp_data->os.data, adata->avp_data->os.len, 0);
					break;
					
				case DIAM_ATTR_LOGIN_LAT_GROUP:
					CONV2_STR(RADIUS_ATTR_LOGIN_LAT_GROUP, adata->avp_data->os.data, adata->avp_data->os.len, 1);
					break;
					
				case DIAM_ATTR_LOGIN_LAT_NODE:
					CONV2_STR(RADIUS_ATTR_LOGIN_LAT_NODE, adata->avp_data->os.data, adata->avp_data->os.len, 1);
					break;
					
				case DIAM_ATTR_LOGIN_LAT_PORT:
					CONV2_STR(RADIUS_ATTR_LOGIN_LAT_PORT, adata->avp_data->os.data, adata->avp_data->os.len, 1);
					break;
					
				case DIAM_ATTR_LOGIN_LAT_SERVICE:
					CONV2_STR(RADIUS_ATTR_LOGIN_LAT_SERVICE, adata->avp_data->os.data, adata->avp_data->os.len, 1);
					break;
					
				case DIAM_ATTR_LOGIN_SERVICE:
					CONV2_32B(RADIUS_ATTR_LOGIN_SERVICE, adata->avp_data->u32);
					break;
					
				case DIAM_ATTR_LOGIN_TCP_PORT:
					CONV2_32B(RADIUS_ATTR_LOGIN_TCP_PORT, adata->avp_data->u32);
					break;
					
			/*
			      -							        If the
        			 Multi-Round-Time-Out AVP is present, the value of the AVP MUST
        			 be inserted in the RADIUS Session-Timeout AVP.

			      o  As described in [NASREQ], if the Result-Code AVP set to
				 DIAMETER_MULTI_ROUND_AUTH and the Multi-Round-Time-Out AVP is
				 present, it is translated to the RADIUS Session-Timeout attribute.
			*/
				case DIAM_ATTR_MULTI_ROUND_TIMEOUT:
					CONV2_32B(RADIUS_ATTR_SESSION_TIMEOUT, adata->avp_data->u32);
					break;
					
				case DIAM_ATTR_NAS_FILTER_RULE:
					/* This is not translatable to RADIUS */
					log_normal("Received Diameter answer with non-translatable NAS-Filter-Rule AVP from '%.*s' (session: '%.*s'), ignoring.\n",
							oh->avp_data->os.len, oh->avp_data->os.data,
							sid->avp_data->os.len, sid->avp_data->os.data);
					handled = 0;
					break;
					
				/* NAS-Identifier is not present in answers */
				/* NAS-IP-Address is not present in answers */
				/* NAS-IPv6-Address is not present in answers */
				/* NAS-Port is not present in answers */
				/* NAS-Port-Id is not present in answers */
				/* NAS-Port-Type is not present in answers */
				
				case DIAM_ATTR_ORIGIN_AAA_PROTOCOL:
					/* We just remove this AVP */
					break;
					
				/* Originating-Line-Info is not present in answers */
				
				case DIAM_ATTR_PASSWORD_RETRY:
					CONV2_32B(RADIUS_ATTR_PASSWORD_RETRY, adata->avp_data->u32);
					break;
				
				case DIAM_ATTR_PORT_LIMIT:
					CONV2_32B(RADIUS_ATTR_PORT_LIMIT, adata->avp_data->u32);
					break;
				
				case DIAM_ATTR_PROMPT:
					CONV2_32B(RADIUS_ATTR_PROMPT, adata->avp_data->u32);
					break;
					
				case DIAM_ATTR_QOS_FILTER_RULE:
					/* This is not translatable to RADIUS */
					log_normal("Received Diameter answer with non-translatable QoS-Filter-Rule AVP from '%.*s' (session: '%.*s'), ignoring.\n",
							oh->avp_data->os.len, oh->avp_data->os.data,
							sid->avp_data->os.len, sid->avp_data->os.data);
					handled = 0;
					break;
					
				/* Re-Auth-Request-Type already handled */
				
				case DIAM_ATTR_REPLY_MESSAGE:
					CONV2_STR(RADIUS_ATTR_REPLY_MESSAGE, adata->avp_data->os.data, adata->avp_data->os.len, 2);
					break;
					
				case DIAM_ATTR_SERVICE_TYPE:
					CONV2_32B(RADIUS_ATTR_SERVICE_TYPE, adata->avp_data->u32);
					break;
				
				case DIAM_ATTR_STATE:
					CONV2_STR(RADIUS_ATTR_STATE, adata->avp_data->os.data, adata->avp_data->os.len, 2);
					break;
					
				case DIAM_ATTR_TUNNELING:
					{
#define CONV2_TUN_STR( _attr_, _data_, _len_, _trunc_)	{					\
	size_t __l = (size_t)(_len_);								\
	size_t __w = (__l > 252) ? 252 : __l;							\
	size_t __off = 0;									\
	if ((_trunc_) == 0) {									\
		CHECK_PARAMS( __l <= 252 );							\
	}											\
	if ((__l > 252) && (_trunc_ == 1)) {							\
		TRACE_DEBUG(FULL, "Attribute truncated!");					\
		__l = 252;									\
	}											\
	buf[0] = tuntag;									\
	memcpy(&buf[1], (_data_), __w);								\
	CHECK_MALLOC(radius_msg_add_attr(*rad_fw, (_attr_), &buf[0], __w + 1));			\
	while (__l -= __w) {									\
		__off += __w;									\
		__w = (__l > 253) ? 253 : __l;							\
		CHECK_MALLOC(radius_msg_add_attr(*rad_fw, (_attr_), (_data_) + __off, __w));	\
	}											\
}

#define CONV2_TUN_32B( _attr_, _data_)	{							\
	uint32_t __v = htonl((uint32_t)(_data_));						\
	         __v |= tuntag << 24;								\
	CHECK_MALLOC(radius_msg_add_attr(*rad_fw, (_attr_), (uint8_t *)&__v, sizeof(__v)));	\
}
						msg_avp_t *inavp, *innext;
						tuntag++;
						CHECK_FCT( msg_browse(avp, MSG_BRW_FIRST_CHILD, &innext, NULL) );
						while (innext) {
							inavp = innext;
							CHECK_FCT( msg_browse(inavp, MSG_BRW_NEXT, &innext, NULL) );
							CHECK_FCT( msg_avp_data ( inavp, &adata ) );
							
							if (adata->avp_flags & AVP_FLAG_VENDOR == 0) {
								switch (adata->avp_code) {
									case DIAM_ATTR_TUNNEL_TYPE:
										CONV2_TUN_32B( RADIUS_ATTR_TUNNEL_TYPE, adata->avp_data->u32);
										break;
										
									case DIAM_ATTR_TUNNEL_MEDIUM_TYPE:
										CONV2_TUN_32B( RADIUS_ATTR_TUNNEL_MEDIUM_TYPE, adata->avp_data->u32);
										break;
										
									case DIAM_ATTR_TUNNEL_CLIENT_ENDPOINT:
										CONV2_TUN_STR(RADIUS_ATTR_TUNNEL_CLIENT_ENDPOINT, adata->avp_data->os.data, adata->avp_data->os.len, 1);
										break;
										
									case DIAM_ATTR_TUNNEL_SERVER_ENDPOINT:
										CONV2_TUN_STR(RADIUS_ATTR_TUNNEL_SERVER_ENDPOINT, adata->avp_data->os.data, adata->avp_data->os.len, 1);
										break;
										
									case DIAM_ATTR_TUNNEL_PREFERENCE:
										CONV2_TUN_32B( RADIUS_ATTR_TUNNEL_PREFERENCE, adata->avp_data->u32);
										break;
										
									case DIAM_ATTR_TUNNEL_CLIENT_AUTH_ID:
										CONV2_TUN_STR(RADIUS_ATTR_TUNNEL_CLIENT_AUTH_ID, adata->avp_data->os.data, adata->avp_data->os.len, 1);
										break;
										
									case DIAM_ATTR_TUNNEL_SERVER_AUTH_ID:
										CONV2_TUN_STR(RADIUS_ATTR_TUNNEL_SERVER_AUTH_ID, adata->avp_data->os.data, adata->avp_data->os.len, 1);
										break;
										
									case DIAM_ATTR_TUNNEL_ASSIGNEMENT_ID:
										CONV2_TUN_STR(RADIUS_ATTR_TUNNEL_ASSIGNEMENT_ID, adata->avp_data->os.data, adata->avp_data->os.len, 1);
										break;
										
									case DIAM_ATTR_TUNNEL_PASSWORD:
										{
											/* This AVP must be encoded for RADIUS (similar to radius_msg_add_attr_user_password)
											    0                   1                   2                   3
											    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
											   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
											   |     Type      |    Length     |     Tag       |   Salt
											   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
											      Salt (cont)  |   String ...
											   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
											*/
											size_t pos;
											int i;
											size_t buflen;
											uint8_t * secret;	/* S */
											size_t secret_len;
											uint8_t hash[16];	/* b(i) */
											const uint8_t *addr[3];
											size_t len[3];
											
											/* We need the request authenticator */
											CHECK_PARAMS(req_auth);

											/* Retrieve the shared secret */
											CHECK_FCT((*cs->rgw_clients_getkey)(cli, &secret, &secret_len));
											
											/* Beginning of the buffer */
											buf[0] = tuntag;
											buf[1] = (uint8_t)(lrand48()); /* A (hi bits) */
											buf[2] = (uint8_t)(lrand48()); /* A (low bits) */
											
											/* The plain text string P */
											CHECK_PARAM(adata->avp_data->os.len < 240);
											buf[3] = adata->avp_data->os.len;
											memcpy(&buf[4], adata->avp_data->os.data, adata->avp_data->os.len);
											memset(&buf[4 + adata->avp_data->os.len], 0, sizeof(buf) - 4 - adata->avp_data->os.len);
											
											/* Initial b1 = MD5(S + R + A) */
											addr[0] = secret;
											len[0] = secret_len;
											addr[1] = req_auth;
											len[1] = 16;
											addr[2] = &buf[1];
											len[2] = 2;
											md5_vector(3, addr, len, hash);
											
											/* Initial c(1) = p(1) xor b(1) */
											for (i = 0; i < 16; i++) {
												buf[i + 3] ^= hash[i];
											}
											pos = 16;
											
											/* loop */
											while (pos < adata->avp_data->os.len + 1) {
												addr[0] = secret;
												len[0] = secret_len;
												addr[1] = &buf[pos - 13];
												len[1] = 16;
												/* b(i) = MD5( S + c(i-1) */
												md5_vector(2, addr, len, hash);
												
												/* c(i) = p(i) xor b(i) */
												for (i = 0; i < 16; i++)
													buf[pos + i + 3] ^= hash[i];

												pos += 16;
											}
											
											CONV2_STR(RADIUS_ATTR_TUNNEL_PASSWORD, &buf[0], pos + 3, 0);
										}
										break;
										
									case DIAM_ATTR_TUNNEL_PRIVATE_GROUP_ID:
										CONV2_TUN_STR(RADIUS_ATTR_TUNNEL_PRIVATE_GROUP_ID, adata->avp_data->os.data, adata->avp_data->os.len, 1);
										break;
									
									default:
										TRACE_DEBUG(FULL, "Ignored unknown AVP inside Tunneling (%d)", adata->avp_code);
								}
							} else {
								TRACE_DEBUG(FULL, "Ignored unknown Vendor AVP inside Tunneling (%d, %d)", adata->avp_vendor, adata->avp_code);
							}
						}
					}
					break;
					
				case DIAM_ATTR_USER_NAME:
					CONV2_STR(RADIUS_ATTR_USER_NAME, adata->avp_data->os.data, adata->avp_data->os.len, 1);
					break;
				
				/* User-Password never present in answers */
					
		/* RFC 4072 (AVP in the order of the EAP Command AVP Table) */
			/*
			      o  Diameter Accounting-EAP-Auth-Method AVPs, if present, are
				 discarded.
			*/
				case DIAM_ATTR_ACCOUNTING_EAP_AUTH_METHOD:
					break;
					
			/*
			      o  Diameter EAP-Master-Session-Key AVP can be translated to the
				 vendor-specific RADIUS MS-MPPE-Recv-Key and MS-MPPE-Send-Key
				 attributes [RFC2548].  The first up to 32 octets of the key is
				 stored into MS-MPPE-Recv-Key, and the next up to 32 octets (if
				 present) are stored into MS-MPPE-Send-Key.  The encryption of this
				 attribute is described in [RFC2548].
			*/
				case DIAM_ATTR_EAP_MASTER_SESSION_KEY:
					{
						uint8_t * secret;	/* S */
						size_t secret_len;
						size_t recv_len, send_len;

						/* We need the request authenticator */
						CHECK_PARAMS(req_auth);

						/* Retrieve the shared secret */
						CHECK_FCT((*cs->rgw_clients_getkey)(cli, &secret, &secret_len));
						
						if (adata->avp_data->os.len != 64) {
							TRACE_DEBUG(INFO, "Received EAP-Master-Session-Key attribute with length %d != 64.\n", adata->avp_data->os.len)
						}
						
						CHECK_PARAMS(adata->avp_data->os.len <= 64);
						recv_len = adata->avp_data->os.len >= 32 ? 32 : adata->avp_data->os.len;
						send_len = adata->avp_data->os.len - recv_len;
						
						if ( ! radius_msg_add_mppe_keys(*rad_fw, req_auth, secret, secret_len, 
								adata->avp_data->os.data + recv_len, send_len,
								adata->avp_data->os.data, recv_len) ) {
							TRACE_DEBUG(INFO, "Error while converting EAP-Master-Session-Key to RADIUS message");
							return ENOMEM;
						}
					}
					break;
				
				case DIAM_ATTR_EAP_KEY_NAME:
					CONV2_STR(RADIUS_ATTR_EAP_KEY_NAME, adata->avp_data->os.data, adata->avp_data->os.len, 1);
					break;
				
			/*
			      o  Diameter EAP-Payload AVP is translated to RADIUS EAP-Message
				 attribute(s).  If necessary, the value is split into multiple
				 RADIUS EAP-Message attributes.
			*/
				case DIAM_ATTR_EAP_PAYLOAD:
					if ( ! radius_msg_add_eap(*rad_fw, adata->avp_data->os.data, adata->avp_data->os.len) ) {
						TRACE_DEBUG(INFO, "Error while converting EAP payload to RADIUS message");
						return ENOMEM;
					}
					break;
					
			/*
			      o  Diameter EAP-Reissued-Payload AVP is translated to a message that
				 contains RADIUS EAP-Message attribute(s), and a RADIUS Error-Cause
				 attribute [RFC3576] with value 202 (decimal), "Invalid EAP Packet
				 (Ignored)" [RFC3579].
			*/
				case DIAM_ATTR_EAP_REISSUED_PAYLOAD:
					if ( ! radius_msg_add_eap(*rad_fw, adata->avp_data->os.data, adata->avp_data->os.len) ) {
						TRACE_DEBUG(INFO, "Error while converting EAP reissued payload to RADIUS message");
						return ENOMEM;
					}
					
					if ( ! radius_msg_add_attr_int32(*rad_fw, RADIUS_ATTR_ERROR_CAUSE, 202) ) {
						TRACE_DEBUG(INFO, "Error while adding Error-Cause attribute in RADIUS message");
						return ENOMEM;
					}
					break;
			
				default:
					/* Leave the AVP in the message for further treatment */
					handled = 0;
			}
		} else {
			/* Vendor-specific AVPs */
			switch (adata->avp_vendor) {
				
				default: /* unknown vendor */
					handled = 0;
			}
		}
		
		if (handled) {
			CHECK_FCT( msg_free( avp, 1 ) );
		}
	}
	
	CHECK_FCT( msg_free( asid, 1 ) );
	CHECK_FCT( msg_free( aoh, 1 ) );
	free(req_auth);

	return 0;
}

int rga_register(int version, waaad_api_t * waaad_api, struct radius_gw_api * api)
{
	TRACE_ENTRY("%d %p %p", version, waaad_api, api);
	CHECK_PARAMS( waaad_api && api );
	
	if (version != RADIUS_GW_API_VER) {
		log_error("ABI version mismatch, please recompile this extension (%s)\n", __FILE__);
		return EINVAL;
	}
	
	/* Required to use the waaad api from this sub-extension: */
	EXTENSION_API_INIT_INTERN( API_MODULE_ALL, "sub_auth", waaad_api );
	
	/* Initialize the radius_gw api callbacks */
	api->rga_conf_parse_cb = auth_conf_parse;
	api->rga_conf_free_cb  = auth_conf_free;
	api->rga_rad_req_cb    = auth_rad_req;
	api->rga_diam_ans_cb   = auth_diam_ans;
	
	/* We're done, we must not initialize any state here since the extension must be re-entrant, but in sample_conf_parse */
	return 0;
}
